
from . import v1
from . import v2

from .v2.__init__ import *

from .__version__ import version_info, __version__

__title__ = "pluralkit"
__author__ = "Madison Landry, Alyx Warner"
__copyright__ = "Copyright 2021-present Madison Landry, Alyx Warner"
